#ifndef WRITE_MODELS_BB_H
#define WRITE_MODELS_BB_H

#include "arch_types.h"

void WriteModels_bb(const char* ArchFile,
                    const char* VEchoFile,
                    const t_arch* arch);

#endif
