/**
 * Jason Luu
 * April 15, 2011
 * pb_type_graph_annotations loads statistical information onto the different nodes/edges of a pb_type_graph.  These statistical informations include delays, capacitance, etc.
 */

#ifndef PB_TYPE_GRAPH_ANNOTATIONS_H
#define PB_TYPE_GRAPH_ANNOTATIONS_H

void load_pb_graph_pin_to_pin_annotations(t_pb_graph_node* pb_graph_node);

#endif
